#include "image.hpp"
#include "state.hpp"

#ifdef __APPLE__
    #include <OpenGL/gl.h>
    #include <OpenGL/gl3.h>
    #include <GLUT/glut.h>
#else
    #include <GL/glew.h>
    #include <GL/glut.h>
#endif

#include <chrono>
#include <cstdlib>
#include <iostream>
#include <string>
#include <arpa/inet.h>
#include <unistd.h>
#include <mutex>
#include <thread>
#include <condition_variable>

using namespace std;
using namespace std::chrono;

uint16_t PORT = 8081;
const uint16_t BUFFER_SIZE = 2048;
int clientSocket = 0;

GLuint imageTexture = 0;
GLuint blankTexture = 0;

GLint screenWidth = 0;
GLint screenHeight = 0;

Image img;
State state;

uint64_t exposureTime = 0;
steady_clock::time_point startTime;
bool displaying = false;
bool completed = false;
bool timing = false;

/**
 * State Message Variables
 **/

bool receivedMessage = false;
string incomingMessage;
string outgoingMessage;

mutex receivedMutex;
condition_variable receivedCondVar;


void actionDisplay () {
    vector<uint64_t> exposure = state.getExposure();
    if (exposure.size() == 0) {
        displaying = true;
        completed = true;
        timing = false;
        startTime = steady_clock::now();
    } else if (exposure.size() == 1) {
        exposureTime = exposure[0];
        displaying = true;
        completed = false;
        timing = true;
        startTime = steady_clock::now();
    } // multi channel case
}

void actionStop () {
    displaying = false;
    completed = true;
    auto currentTime = steady_clock::now();
    auto elapsedTime = duration_cast<milliseconds>(currentTime - startTime).count();
    cout << "{ \"exposed\" : " << elapsedTime << ", \"exposure\" : null }" << endl;
}

void postAction () {
    if (state.isError()) {
        outgoingMessage = state.createMessage(false);
    } else {
        outgoingMessage = state.createMessage(true);
    }
    const char* msg = outgoingMessage.c_str();
    int bytesSent = send(clientSocket, msg, strlen(msg), 0);
    if (bytesSent < 0) {
        cerr << "{ \"error\" : \"Error sending ACK to client\" }" << endl;
    } else {
        cout << "{ \"sent\" : " << outgoingMessage << " }" << endl;
    }
}

void actionLoad () {
    auto localStartTime = steady_clock::now();
    string imagePath = state.getImage();
    vector<uint64_t> position = state.getPosition();
    Mat image = img.loadImage(imagePath, position[0], position[1], position[2], position[3]);
    glGenTextures(1, &imageTexture);
    glBindTexture(GL_TEXTURE_2D, imageTexture);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    glTexImage2D(GL_TEXTURE_2D,          // Type of texture
                    0,                   // Pyramid level (for mip-mapping) - 0 is the top level
                    GL_RGB,              // Internal colour format to convert to
                    image.cols,          // Image width 
                    image.rows,          // Image height
                    0,                   // Border width in pixels (can either be 1 or 0)
                    GL_RGB,              // Input image format (i.e. GL_RGB, GL_RGBA, GL_BGR etc.)
                    GL_UNSIGNED_BYTE,    // Image data type
                    image.ptr());        // The actual image data itself
    auto localCurrentTime = steady_clock::now();
    auto localElapsedTime = duration_cast<milliseconds>(localCurrentTime - localStartTime).count();
    cout << "{ \"load_time\" : " << localElapsedTime << " }" << endl;
    completed = true;
}

void loadBlank () {
    Mat blank = img.getBlank();
    glGenTextures(1, &blankTexture);
    glBindTexture(GL_TEXTURE_2D, blankTexture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    glTexImage2D(GL_TEXTURE_2D,          // Type of texture
                    0,                   // Pyramid level (for mip-mapping) - 0 is the top level
                    GL_RGB,              // Internal colour format to convert to
                    blank.cols,          // Image width 
                    blank.rows,          // Image height
                    0,                   // Border width in pixels (can either be 1 or 0)
                    GL_RGB,              // Input image format (i.e. GL_RGB, GL_RGBA, GL_BGR etc.)
                    GL_UNSIGNED_BYTE,    // Image data type
                    blank.ptr());        // The actual image data itself
}

void initImageTexture() {
    glGenTextures(1, &imageTexture);
}

void display () {
    if (state.isActive()) {
        if (state.getAction() == LOAD) {
            actionLoad();
        } else if (state.getAction() == DISPLAY) {
            actionDisplay();
        } else if (state.getAction() == STOP) {
            actionStop();
        }
        state.setInactive();
    }

    if (completed) {
        postAction();
        completed = false;
    }

    if (displaying && timing) {
        auto currentTime = steady_clock::now();
        auto elapsedTime = duration_cast<milliseconds>(currentTime - startTime).count();
        if (elapsedTime >= exposureTime) {
            cout << "{ \"exposed\" : " << elapsedTime << ", \"exposure\" : " <<  exposureTime << " }" << endl;
            displaying = false;
            timing = false;
            completed = true;
        }
    }

    glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_TEXTURE_2D);
    glViewport(0, 0, screenWidth, screenHeight);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

    if (displaying) {
        glBindTexture(GL_TEXTURE_2D, imageTexture);
    } else {
        glBindTexture(GL_TEXTURE_2D, blankTexture);
    }

    glBegin(GL_QUADS);  // front face
        glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f, 0.0f);  //bottom right
        glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, 0.0f);  //top right
        glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, 0.0f);  //top left
        glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, 0.0f);  //bottom left
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, 0);
    glutSwapBuffers();

  //glFlush();
}

void timer(int value) {
    glutPostRedisplay();
    glutTimerFunc(1, timer, 0);
}

/**
 * 
 **/

void processMessage () {
    if (receivedMessage) {
        state.receiveMessage(incomingMessage);
        receivedMessage = false;
    }
}

void handleTCPConnection(int clientSocket) {
    char buffer[BUFFER_SIZE];
    memset(buffer, 0, BUFFER_SIZE);

    while (true) {
        int bytesRead = recv(clientSocket, buffer, BUFFER_SIZE - 1, 0);
        if (bytesRead <= 0) {
            cerr << "{ \"error\" : \"Client disconnected or errored out\" }" << endl;
            break;
        }

        string localIncomingMessage(buffer, bytesRead);

        {
            lock_guard<mutex> lock(receivedMutex);
            receivedMessage = true;
            incomingMessage = localIncomingMessage;
            receivedCondVar.notify_all();
        }

        processMessage();

        memset(buffer, 0, BUFFER_SIZE);
    }
}

void runTCPServer (int serverSocket) {
    while (true) {
        struct sockaddr_in clientAddr;
        socklen_t clientAddrLen = sizeof(clientAddr);

        clientSocket = accept(serverSocket, (struct sockaddr*)&clientAddr, &clientAddrLen);
        if (clientSocket < 0) {
            cerr << "{ \"error\" : \"Error accepting client connection\" }" << endl;
            continue;
        }

        cout << "{ \"client\" : \"CONNECTED\" }" << endl;

        handleTCPConnection(clientSocket);
    }
}

uint8_t handleArgs (int argc, char** argv) {
    if (argc > 1) {
        screenWidth = atoi(argv[1]);
        cout << "{ \"screen_width\" : " << screenWidth << " }" << endl;
    } else {
        cerr << "{ \"error\" : \"Please provide screen width as first argument\" }" << endl;
        return 1;
    }
    if (argc > 2) {
        screenHeight = atoi(argv[2]);
        cout << "{ \"screen_height\" : " << screenHeight << " } " << endl;
    } else {
        cerr << "{ \"error\" : \"Please provide screen height as second argument\" }" << endl;
        return 2;
    }
    if (argc > 3) {
        PORT = (uint16_t) atoi(argv[3]);
        cout << "{ \"port\" : " << PORT << " }" << endl;
    }
    return 0;
}

int main (int argc, char** argv) {
    uint8_t argsRet = handleArgs(argc, argv);

    if (argsRet > 0) {
        return argsRet;
    }

    img.setDimensions( (uint16_t) screenWidth, (uint16_t) screenHeight);

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_SINGLE);
    glutCreateWindow("filmout_display");
    glutSetCursor(GLUT_CURSOR_NONE);
    glutFullScreen();
    glutDisplayFunc(display);
    glutTimerFunc(0, timer, 0);
    loadBlank();
    initImageTexture();

    int serverSocket = socket(AF_INET, SOCK_STREAM, 0);
    if (serverSocket < 0) {
        cerr << "{ \"error\" : \"Error creating socket\" }" << endl;
        return 1;
    }

    struct sockaddr_in serverAddr;
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_addr.s_addr = INADDR_ANY;
    serverAddr.sin_port = htons(PORT);

    int bindResult = ::bind(serverSocket, reinterpret_cast<struct sockaddr*>(&serverAddr), sizeof(serverAddr));
    if (bindResult < 0) {
        cerr << "{ \"error\" : \"Error binding socket\" }" << endl;
        close(serverSocket);
        return 1;
    }

    if (listen(serverSocket, 5) < 0) {
        cerr << "{ \"error\" : \"Error listening on socket\" }" << endl;
        return 1;
    }

    cout << "{ \"listening\" : true,  \"port\" : " << PORT << " }" << endl;

    thread serverThread(runTCPServer, serverSocket);

    glutMainLoop();

    serverThread.join();

    close(serverSocket);

    return 0;
}